@@ -6,7 +6,13 @@ class Field:
     inherited_from:   str
     visibility: str
     name:       str
-    line_num:   tuple # (start, end)
+    line_num:   tuple  # (start, end)
+
+
+@dataclass
+class Modifier:
+    name: str
+    visibility: str
 
 
 @dataclass
@@ -24,28 +30,29 @@ class Function:
     line_num: tuple  # (start, end)
 
 
+@dataclass
+class Event:
+    name: str
+    signature:  str
+    anonymous: bool
+    line_num: tuple
+    raw: str
+
+
 @dataclass
 class ContractData:
     abstract:       bool
     name:           str
     kind:           str
     base_contracts: List[int]
-    fields:         list
-    functions:      list
-    modifiers:      list
-    line_num:       tuple # (start, end)
-    contract_id: int # unique id in ast per solc compilation
-
-@dataclass
-class Modifier:
-    name: str
-    visibility: str
+    fields:         List[Field]
+    functions:      List[Function]
+    modifiers:      List[Modifier]
+    line_num:       tuple  # (start, end)
+    contract_id:    int    # unique id in ast per solc compilation
+    events:         List[Event]
 
 
 
 
 
-# [Function(signature='',           return_signature='',   name='receive',  modifiers=[], type='receive'),
-#  Function(signature='receive()',  return_signature='()', name='receive',  modifiers=[], type='function'),
-#  Function(signature='fallback()', return_signature='()', name='fallback', modifiers=[], type='function'),
-#  Function(signature='',           return_signature='',   name='fallback', modifiers=[], type='fallback')]
\ No newline at end of file
