@@ -9,14 +9,15 @@ import json
 import os
 import re
 from typing import Collection, Dict, Optional, List, Any, Tuple, Union
-from functools import cached_property, cache
+from functools import cached_property, cache, reduce
+from Crypto.Hash import keccak
 
 try:
-    from fields import Field, Function, ContractData, Modifier
+    from fields import Field, Function, ContractData, Modifier, Event
     from version_cfg import v_keys
     import consts
 except:
-    from solc_json_parser.fields import Field, Function, ContractData, Modifier
+    from solc_json_parser.fields import Field, Function, ContractData, Modifier, Event
     from solc_json_parser.version_cfg import v_keys
     import solc_json_parser.consts
 
@@ -46,6 +47,10 @@ DEPLOY_START_OPCODES = [
     ],
 ]
 
+def keccak256(s: str) -> str:
+    k = keccak.new(digest_bits=256)
+    k.update(s.encode())
+    return k.hexdigest()
 
 def get_by_index(lst: Union[List, Tuple], idx: int):
     '''Get by index from a list, returns None if the index is out of range '''
@@ -71,7 +76,8 @@ def get_in(d, key: Any, *nkeys) -> Any:
 
 def get_candidates():
     '''
-    Returns a cached list of solc versions available for install
+    Returns a cached list of solc versions available for install,
+    version list is sorted in ascending order
     '''
     global INSTALLABLE_VERSION
     if INSTALLABLE_VERSION:
@@ -80,6 +86,7 @@ def get_candidates():
         INSTALLABLE_VERSION = sorted(solcx.get_installable_solc_versions())
         return INSTALLABLE_VERSION
 
+
 def select_available_version(version_str: str, install=False) -> Optional[str]:
     '''Switch to current or the next semantic version available to use. Returns the version selected.'''
     version = Version(version_str)
@@ -98,6 +105,7 @@ def select_available_version(version_str: str, install=False) -> Optional[str]:
 
     return ver
 
+
 def version_str_from_line(line) -> Optional[str]:
     '''
     Extract solc version string from input line
@@ -108,6 +116,7 @@ def version_str_from_line(line) -> Optional[str]:
         return re.sub(r'(\.0+)', '.0', ver)
     return None
 
+
 def version_str_from_source(source_or_source_file: str) -> Optional[str]:
     inputs = source_or_source_file.split('\n') if '\n' in source_or_source_file else open(source_or_source_file, 'r')
 
@@ -139,25 +148,33 @@ def detect_solc_version(source_or_source_file: str) -> Optional[str]:
     return str(next(spec.filter(candidates)))
 
 
-def get_line_number_range(start_index:int, offset:int, source_code:str):
-    start_line = source_code[:start_index].count('\n') + 1
-    end_line   = start_line + source_code[start_index:start_index + offset].count('\n')
-    return start_line, end_line
-
 def symbols_to_ids_from_ast_v8(ast: Dict[Any, Any]) -> Dict[str, int]:
     syms = [c['ast']['exportedSymbols'] for c in ast.values()]
     return {k: v[0] for m in syms for k, v in m.items()}
 
+
 def symbols_to_ids_from_ast_v7(ast: Dict[Any, Any]) -> Dict[str, int]:
     syms = [c['ast']['attributes']['exportedSymbols'] for c in ast.values()]
     return {k: v[0] for m in syms for k, v in m.items()}
 
 
+def get_increased_version(current_version: str) -> str:
+    """
+    :param current_version:
+    :return: next solc valid version, e.g. 0.5.10 -> 0.5.11
+    """
+    # convert current version str to Version object
+    current_version_obj = Version(current_version)
+    next_version = select_available_version(str(current_version_obj.next_patch()), install=True)
+    if not next_version:
+        raise SolidityAstError(f'No next version available for {current_version}')
+    return str(next_version)
+
+
 class SolidityAstError(ValueError):
     pass
 
 
-
 class SolidityAst():
 
     FIELD_VISIBILITY_ALL = frozenset(('default', 'internal', 'public', 'private'))
@@ -166,18 +183,71 @@ class SolidityAst():
     FUNC_VISIBILITY_ALL = frozenset(('external', 'private', 'internal', 'public'))
     FUNC_VISIBILITY_NON_PRIVATE = frozenset(('external', 'internal', 'public'))
 
-    def __init__(self, contract_source_path: str, version=None):
+    def __init__(self, contract_source_path: str, version=None, retry_num=None, solc_options={}):
+        '''
+    Compile the input contract and create a SolidityAst object.
+
+    Parameters:
+    contract_source_path: str, required
+        a path to the solidity source file or source code as string.
+    version:  str, optional
+        solc version to use for compile the contract. If not provided will use auto detected solc version.
+        Note that SolidityAst will try to compile the contract starting from the lowest detected solc version first, increase the solc version each time when compilation fails.
+    retry_num: int, optional
+        Maximum number of solc versions to try to compile the contract
+
+
+    solc_options: Dict, optional
+    The optionsl passed to the solc compiler, the following options are supports:
+    overwrite : bool, optional
+        Overwrite existing files (used in combination with `output_dir`)
+    evm_version: str, optional
+        Select the desired EVM version. Valid options depend on the `solc` version.
+    revert_strings : List | str, optional
+        Strip revert (and require) reason strings or add additional debugging
+        information.
+    metadata_hash : str, optional
+        Choose hash method for the bytecode metadata or disable it.
+    metadata_literal : bool, optional
+        Store referenced sources as literal data in the metadata output.
+    optimize : bool, optional
+        Enable bytecode optimizer.
+    optimize_runs : int, optional
+        Set for how many contract runs to optimize. Lower values will optimize
+        more for initial deployment cost, higher values will optimize more for
+        high-frequency usage.
+    optimize_yul: bool, optional
+        Enable the yul optimizer.
+    no_optimize_yul : bool, optional
+        Disable the yul optimizer.
+    yul_optimizations : int, optional
+        Force yul optimizer to use the specified sequence of optimization steps
+        instead of the built-in one.
+    solc_binary : str | Path, optional
+        Path of the `solc` binary to use. If not given, the currently active
+        version is used (as set by `solcx.set_solc_version`)
+        '''
         if '\n' in contract_source_path:
             self.source = contract_source_path
             self.file_path = None
             self.root_path = None
+            self.compile_type = 'source'
         else:
+            self.compile_type = 'file'
             self.file_path = contract_source_path
-            self.root_path = os.path.dirname(self.file_path)
+            self.file_path = os.path.abspath(contract_source_path)
+            self.root_path = os.path.dirname(contract_source_path)
             with open(contract_source_path, 'r') as f:
                 self.source = f.read()
 
-        self.exact_version: str   =  version or detect_solc_version(self.source) or consts.DEFAULT_SOLC_VERSION
+        self.original_compilation_output :Optional[Dict] = None
+        self.solc_options = solc_options
+        self.import_remappings = solc_options.get('import_remappings')
+        base_path = solc_options.get('base_path')
+        self.base_path = os.path.abspath(base_path) if base_path else None
+        self.allow_paths = solc_options.get('allow_paths')
+        self.retry_num = retry_num or 0
+        self.exact_version: str   = version or detect_solc_version(self.source) or consts.DEFAULT_SOLC_VERSION
         self.version_key: str     = self._get_version_key()
         self.keys: addict.Dict    = v_keys[self.version_key]
         self.exported_symbols: Dict[str, int] = {} # contract name -> id mapping, to be determined in _parse()
@@ -207,27 +277,34 @@ class SolidityAst():
             base_contracts.append(base_contract['baseName']['referencedDeclaration'])
         return base_contracts
 
+    def get_raw_from_src(self, node):
+        line_number_range_raw = list(map(int, node.get('src').split(':')))
+        start, offset, source_file_idx = line_number_range_raw
+        line_number_range, source = self.get_line_number_range_and_source(line_number_range_raw)
+        raw = source.encode()[start: start+offset].decode()
+        return raw, line_number_range
+
+    def get_signature(self, function_name, parameters, kind='function') -> str:
+        if kind in ['function', 'constructor']:
+            return ''
+
+        signature = "" + function_name + "("
+        param_type_str = ""
+        if self.v8:
+            for param in parameters['parameters']:
+                param_type_str += param['typeDescriptions']['typeString'] + ", "
+        else:  # v4, v5, v6, v7
+            for param in parameters['children']:
+                param_type_str += param['attributes']['type'] + ", "
+
+        param_type_str = param_type_str[:-2]  # remove the last ", "
+        signature += param_type_str + ")"
+        return signature
+
     def _process_function(self, node: Dict) -> Function:
-        def _get_signature(function_name, parameters, kind='function') -> str:
-            if kind == 'fallback' or kind == 'receive':
-                return ''
-
-            signature = "" + function_name + "("
-            param_type_str = ""
-            if self.version_key == "v8":
-                for param in parameters['parameters']:
-                    param_type_str += param['typeDescriptions']['typeString'] + ", "
-            else:  # v4, v5, v6, v7
-                for param in parameters['children']:
-                    param_type_str += param['attributes']['type'] + ", "
-
-            param_type_str = param_type_str[:-2] # remove the last ", "
-            signature += param_type_str + ")"
-            return signature
-
         def _get_modifiers(node: Dict) -> List[str]:
             modifiers = []
-            if self.version_key == "v8":
+            if self.v8:
                 for modifier in node['modifiers']:
                     modifiers.append(modifier['modifierName']['name'])
                 return modifiers
@@ -242,12 +319,8 @@ class SolidityAst():
             return modifiers
 
         # line number range is the same for all versions
-        line_number_range_raw = list(map(int, node.get('src').split(':')))
-        line_number_range = get_line_number_range(start_index=line_number_range_raw[0], offset=line_number_range_raw[1], source_code=self.source)
-        start, offset = line_number_range_raw[0], line_number_range_raw[1]
-        raw = self.source.encode()[start: start+offset].decode()
-
-        if self.version_key == "v8":
+        raw, line_number_range = self.get_raw_from_src(node)
+        if self.v8:
             parameters = node.get('parameters')
             return_type = node.get('returnParameters')
             modifier_nodes = node
@@ -266,7 +339,6 @@ class SolidityAst():
                 modifier_nodes = {"modifiers": []}
             node = node.get("attributes") # get attributes, the structure is different
 
-
         assert parameters  is not None
         assert return_type is not None
         visibility = node.get('visibility')
@@ -284,17 +356,16 @@ class SolidityAst():
         inherited_from = ""
         abstract  = not node.get('implemented')
         modifiers = _get_modifiers(modifier_nodes)
-        func_kind = node.get('kind')
+        kind = node.get('kind')
         state_mutability = node.get('stateMutability')
 
-        signature = _get_signature(name, parameters, func_kind)
-        return_signature = _get_signature("", return_type, func_kind)
+        signature = self.get_signature(name, parameters, kind)
+        return_signature = self.get_signature("", return_type, kind)
         return Function(inherited_from=inherited_from, abstract=abstract, visibility=visibility, raw=raw,
-                        signature=signature, name=name, return_signature=return_signature, kind=func_kind,
+                        signature=signature, name=name, return_signature=return_signature, kind=kind,
                         modifiers=modifiers, line_num=line_number_range, state_mutability=state_mutability)
 
-
-    def get_yul_lines(self, contract_name: str, deploy: bool)->List[str]:
+    def get_yul_lines(self, contract_name: str, deploy: bool) -> List[str]:
         if not self.v8:
             return []
 
@@ -303,7 +374,6 @@ class SolidityAst():
         else:
             return self.solc_json_ast.get(contract_name).get('generated-sources-runtime')[0]['contents'].split("\n")
 
-
     @cached_property
     def v8(self):
         return self.version_key == "v8"
@@ -311,9 +381,9 @@ class SolidityAst():
     def _process_field(self, node: Dict) -> Field:
         # line number range is the same for all versions
         line_number_range_raw = list(map(int, node.get('src').split(':')))
-        line_number_range = get_line_number_range(start_index=line_number_range_raw[0], offset=line_number_range_raw[1], source_code=self.source)
+        line_number_range, _ = self.get_line_number_range_and_source(line_number_range_raw)
 
-        if self.version_key == "v8":
+        if self.v8:
             pass
         else:  # v4, v5, v6, v7
             node = node.get("attributes")
@@ -322,8 +392,26 @@ class SolidityAst():
         inherited_from = ""
         return Field(inherited_from=inherited_from, visibility=visibility, name=name, line_num=line_number_range)
 
+    def _process_event(self, node: Dict) -> Event:
+        raw, line_number_range = self.get_raw_from_src(node)
+        if self.v8:
+            parameters = node.get('parameters')
+        else:  # v4, v5, v6, v7
+            parameters = None
+            for i in range(len(node.get('children'))):
+                if node.get('children')[i].get('name') == "ParameterList":
+                    parameters = node.get('children')[i]
+                    break
+            node = node.get("attributes")
+
+        name = node.get('name')
+        anonymous = node.get('anonymous')
+
+        signature = self.get_signature(name, parameters, "event")
+        return Event(raw=raw, name=name, anonymous=anonymous, line_num=line_number_range, signature=signature)
+
     def _process_modifier(self, node: Dict) -> Modifier:
-        if self.version_key == "v8":
+        if self.v8:
             pass
         else:  # v4, v5, v6, v7
             node = node.get("attributes")
@@ -334,10 +422,10 @@ class SolidityAst():
     def _get_contract_meta_data(self, node: Dict) -> tuple:
         # line number range is the same for all versions
         line_number_range_raw = list(map(int, node.get('src').split(':')))
-        line_number_range = get_line_number_range(start_index=line_number_range_raw[0], offset=line_number_range_raw[1], source_code=self.source)
+        line_number_range, _ = self.get_line_number_range_and_source(line_number_range_raw)
         contract_id = node.get('id')
 
-        if self.version_key == "v8":
+        if self.v8:
             pass  # do nothing
         else:  # v4, v5, v6, v7
             node = node.get("attributes")
@@ -366,6 +454,7 @@ class SolidityAst():
         functions = []
         fields = []
         modifiers = []
+        events = []
         keys = self.keys
         for node in node.get(keys.children):
             if node[keys.name] == "FunctionDefinition":
@@ -374,11 +463,13 @@ class SolidityAst():
                 fields.append(self._process_field(node))
             elif node[keys.name] == "ModifierDefinition":
                 modifiers.append(self._process_modifier(node))
+            elif node[keys.name] == "EventDefinition":
+                events.append(self._process_event(node))
             else:
                 # not implemented for other types
                 pass
 
-        return ContractData(is_abstract, contract_name, contract_kind, base_contracts, fields, functions, modifiers, line_number_range, contract_id)
+        return ContractData(is_abstract, contract_name, contract_kind, base_contracts, fields, functions, modifiers, line_number_range, contract_id, events)
 
     def _parse(self) -> Dict:
         def _add_inherited_function_fields(data_dict: Dict[int, ContractData]):
@@ -396,7 +487,7 @@ class SolidityAst():
                             new_function.inherited_from = base_contract_name
                             contract.functions.append(new_function)
 
-        # self.save_solc_ast_json("Storage_solc_ast_no_abstractBugC")
+        # self.save_solc_ast_json("multi_file5.0")
         # if there are n contracts in the same file, there will be n keys in the json,
         # but we only need the first one[0], because it contains all the contracts, and the rest are the same
         # ast = self.solc_json_ast.get(list(self.solc_json_ast.keys())[0]).get('ast')
@@ -440,17 +531,33 @@ class SolidityAst():
             return None
 
     def compile(self):
-        # print("downloading compiler, version: ", self.exact_version)
         current_working_dir = os.getcwd()
         try:
             solcx.install_solc(self.exact_version)
             solcx.set_solc_version(self.exact_version)
             if self.root_path:
                 os.chdir(self.root_path)
-            out = solcx.compile_source(self.source, output_values=self.solc_compile_outputs, solc_version=self.exact_version)
+                self.root_path = os.getcwd()
+
+            compiler_options = dict(self.solc_options)
+            overwritten_options = dict(base_path=self.base_path,
+                                       import_remappings=self.import_remappings,
+                                       output_values=self.solc_compile_outputs,
+                                       solc_version=self.exact_version)
+            compiler_options.update(overwritten_options)
+            if self.compile_type == "file":
+                out = solcx.compile_files(self.file_path, **compiler_options)
+            else:
+                out = solcx.compile_source(self.source, **compiler_options)
+            self.original_compilation_output = out
             self.solc_json_ast = {k.split(':')[-1]: v for k, v in out.items()}
         except Exception as e:
-            raise SolidityAstError(f"Compile failed: {e}")
+            if self.retry_num > 0:
+                self.retry_num -= 1
+                self.exact_version = get_increased_version(self.exact_version)
+                self.compile()
+            else:
+                raise SolidityAstError(f"Compile failed with solc version {self.exact_version}, err msg: {e}")
         finally:
             os.chdir(current_working_dir)
 
@@ -462,7 +569,6 @@ class SolidityAst():
         with open(f'{PARSED_JSON}/{name}.json', 'w') as f:
             json.dump(self.contracts_dict, f, default=lambda obj: obj.__dict__, indent=4)
 
-
     def all_contracts(self) -> List[ContractData]:
         # dict to list
         return list(self.contracts_dict.values())
@@ -542,7 +648,8 @@ class SolidityAst():
     def functions_in_contract(self, contract: ContractData,
                               name_only: bool = False,
                               function_visibility: Optional[frozenset] = None,
-                              check_base_contract=True) -> List[Function]:
+                              check_base_contract=True) -> List[Union[Function, str]]:
+        # filter and return all functions for a given contract object of type `ContractData`
 
         # by default, base contract's functions are included
         # different from fields, we don't check parent function visibility
@@ -561,15 +668,7 @@ class SolidityAst():
                                       name_only: bool = False,
                                       function_visibility: Optional[frozenset] = None,
                                       check_base_contract=True) -> List[Any]:
-        # fns = self.contract_by_name(contract_name).functions
-        # if check_base_contract:
-        #     pass # do nothing
-        # else:
-        #     fns = [fn for fn in fns if fn.inherited_from == '']
-        #
-        # if name_only:
-        #     return [fn.name for fn in fns]
-
+        # return all functions for a given "contract name"(str)
         contract = self.contract_by_name(contract_name)
         return self.functions_in_contract(contract, name_only, function_visibility, check_base_contract)
 
@@ -581,10 +680,29 @@ class SolidityAst():
         return fns
 
     def function_by_name(self, contract_name: str, function_name: str) -> Function:
+        """return a function for a given "contract name"(str) and "function name"(str)"""
         contract = self.contract_by_name(contract_name)
         funcs    = self.functions_in_contract(contract)
         return next(fn for fn in funcs if fn.name == function_name)
 
+    def events_in_contract(self, contract: ContractData, name_only: bool = False) -> List[Union[str, Event]]:
+        """return all events for a given contract object of type `ContractData`"""
+        events = contract.events
+        if name_only:
+            return [n.name for n in events]
+        return events
+
+    def events_in_contract_by_name(self, contract_name: str, name_only: bool = False) -> List[Any]:
+        """return all events for a given "contract name"(str)"""
+        contract = self.contract_by_name(contract_name)
+        return self.events_in_contract(contract, name_only)
+
+    def event_by_name(self, contract_name: str, event_name: str) -> Event:
+        """return an event for a given "contract name"(str) and "event name"(str)"""
+        contract = self.contract_by_name(contract_name)
+        events    = self.events_in_contract(contract)
+        return next(ev for ev in events if ev.name == event_name)
+
     def all_libraries(self) -> List[ContractData]:
         return [contract for contract in self.all_contracts() if contract.kind == "library"]
 
@@ -613,7 +731,6 @@ class SolidityAst():
             return opcodes[offset - len(deploy_start_sequence) + 1:]
         return SolidityAst.__skip_deploys(opcodes, deploy_sig_idx+1)
 
-
     @staticmethod
     def __record_jumps(opcode: str, code: list[Dict[str, Any]], idx: int, pc: int, seen_targets: set[int]) -> set[int]:
         if opcode == 'JUMPI':
@@ -636,10 +753,7 @@ class SolidityAst():
         asm_data = contract.get('asm').get('.code') if deploy else contract.get('asm').get('.data')
         # deploys = contract.get('asm').get('.code')
         opcodes = contract.get('opcodes').split()
-        if self.v8:
-            source_list = contract.get('asm').get('sourceList')
-        else:
-            source_list = [f"{key_name}.sol" for key_name in combined_json.keys()]
+        source_list = self.get_source_list()
 
         if not deploy:
             opcodes = SolidityAst.__skip_deploys(opcodes)
@@ -696,10 +810,37 @@ class SolidityAst():
         pc2idx = {v: k for k, v in idx2pc.items()}
         return dict(code=code, pc2idx=pc2idx, source_list=source_list, seen_targets=seen_targets)
 
+    @cache
+    def get_source_list(self):
+        source_list = []
+        idx2path = {}
+        for contract_name in self.solc_json_ast.keys():
+            absolute_path = self.source_path_by_contract(contract_name)
+            idx = get_in(self.solc_json_ast, contract_name, 'ast', 'src').split(':')[-1]
+            idx2path[int(idx)] = absolute_path
+        sort_keys = sorted(idx2path.keys())
+        for idx in sort_keys:
+            source_list.append(idx2path[idx])
+        return source_list
+
+    def get_line_number_range_and_source(self, line_number_range_raw: list):
+        start_index, offset, source_file_idx = line_number_range_raw
+        source_list = self.get_source_list()
+        source_path = self.__source_path_from_source_list(source_list, source_file_idx)
+        source_code = self.__source_code_from_source_path(source_path)
+        start_line = source_code[:start_index].count('\n') + 1
+        end_line = start_line + source_code[start_index:start_index + offset].count('\n')
+        return (start_line, end_line), source_code
 
     def source_path_by_contract(self, contract_name) -> Optional[str]:
-        path = get_in(self.solc_json_ast, contract_name, 'ast', 'attributes', 'absolutePath')
-        return None if (not path) or path == '<stdin>' else os.path.join(self.root_path, path)
+        path = None
+        if self.v8:
+            path = get_in(self.solc_json_ast, contract_name, 'ast', 'absolutePath')
+        else:
+            path = get_in(self.solc_json_ast, contract_name, 'ast', 'attributes', 'absolutePath')
+
+        base_path = self.base_path or self.root_path
+        return None if (not path) or path == '<stdin>' else os.path.join(base_path, path)
 
     def source_by_lines(self, contract_name: str, line_start: int, line_end: int) -> str:
         '''Get source code by contract name and line numbers, line numbers are zero indexed'''
@@ -726,14 +867,51 @@ class SolidityAst():
 
     def coverage(self, contract_name: str, pcs: Collection[int]) -> float:
         all_pcs = self.all_pcs(contract_name)
-        return len(set(pcs)) /  len(all_pcs) * 100 if all_pcs else 0
-
+        return len(set(pcs)) / len(all_pcs) * 100 if all_pcs else 0
 
     def __source_path_from_source_list(self, source_list: Optional[List[str]], source_index: Optional[int]) -> Optional[str]:
         if source_index is not None and source_list:
             return source_list[source_index]
         return None
-    
+
+    def __source_code_from_source_path(self, source_path):
+        base_path = self.base_path or self.root_path
+        if base_path and source_path and source_path != '<stdin>':
+            source_path = os.path.join(base_path, source_path)
+            with open(source_path, 'r') as f:
+                source_code = f.read()
+        else:
+            source_code = self.source
+        return source_code
+
+    @staticmethod
+    @cache
+    def parse_src_mapping(srcmap: str):
+        def _reduce_fn(accumulator, current_value):
+            last, *tlist = accumulator
+            return [
+                {
+                    's': int(current_value['s'] or last['s']),
+                    'l': int(current_value['l'] or last['l']),
+                    'f': int(current_value['f'] or last['f']),
+                },
+                last,
+                *tlist
+            ]
+
+        parsed = srcmap.split(";")
+        parsed = [l.split(':') for l in parsed]
+        t = []
+        for l in parsed:
+            if len(l) >= 3:
+                t.append(l[:3])
+            else:
+                t.append(l + [None] * (3 - len(l)))
+        parsed = [{'s': s if s != "" else None, 'l': l, 'f': f} for s, l, f in t]
+        parsed = reduce(_reduce_fn, parsed, [{}])
+        parsed = list(reversed(parsed[:-1]))
+        return parsed
+
     def source_by_pc(self, contract_name: str, pc: int, deploy=False) -> Dict[str, Any]:
         '''
         Get source code by program counter:
@@ -741,23 +919,26 @@ class SolidityAst():
         - `deploy`: set to true to search in deploy opcodes
         '''
         code, pc2idx, source_list = itemgetter('code', 'pc2idx', 'source_list')(self.__parse_asm_data(contract_name, deploy=deploy))
-        part = code[pc2idx[pc]]
+        pc_idx = pc2idx.get(pc)
+        part = code[pc_idx]
+        if part.get('source') is not None:
+            source_idx = part['source']
+        else:
+            src_mapping = self.solc_json_ast[contract_name]['srcmap-runtime']
+            parsed_mapping = self.parse_src_mapping(src_mapping)
+            mapping_idx = list(pc2idx.values()).index(pc_idx)
+            source_idx = parsed_mapping[mapping_idx].get('f')
 
         begin, end = itemgetter('begin', 'end')(part)
-        source_idx = part.get('source')
+        source_idx = source_idx if source_idx is not None else list(self.solc_json_ast.keys()).index(contract_name)
         source_path = self.__source_path_from_source_list(source_list, source_idx)
-        if self.root_path and source_path and source_path != '<stdin>':
-            source_path = os.path.join(self.root_path, source_path)
-            with open(source_path, 'r') as f:
-                source_code = f.read()
-        else:
-            source_code = self.source
+        source_code = self.__source_code_from_source_path(source_path)
 
-        source = get_in(part, 'source') or 0 # WARN: assuming yul source is has index 1, not true for multi-source file contract
-
-        if source == 1 and self.v8 and not deploy:
+        source = source_idx # WARN: assuming yul source is has index 1, not true for multi-source file contract
+        yul_index = len(source_list) - 1
+        if source == yul_index and self.v8 and not deploy:
             combined_source = self.solc_json_ast[contract_name]['generated-sources-runtime'][0]['contents']
-        elif source == 1 and self.v8 and deploy:
+        elif source == yul_index and self.v8 and deploy:
             combined_source = self.solc_json_ast[contract_name]['generated-sources'][0]['contents']
         else:
             combined_source = source_code
@@ -769,6 +950,21 @@ class SolidityAst():
                     source_as_bytes[:end].decode().count('\n') + 1)
         return dict(pc=pc, fragment=fragment, begin=begin, end=end, linenums=linenums, source_idx=source, source_path=(source_path or self.file_path))
 
+    def get_any(self, *keys) -> Any:
+        '''Get any value by keys from the original compiled ast data'''
+        return get_in(self.original_compilation_output, *keys)
+
+    def get_deploy_bin_by_contract_name(self, contract_name: str) -> Optional[str]:
+        return get_in(self.solc_json_ast, contract_name, 'bin')
+
+    def qualified_name_from_hash(self, hsh: str)->str:
+        '''Get fully qualified contract name from 34 character hash'''
+        return next(full_name for full_name in self.original_compilation_output.keys() if keccak256(full_name)[:34] == hsh)
+
+    def get_deploy_bin_by_hash(self, hsh: str) -> Optional[str]:
+        '''Get deployment binary by hash of fully qualified contract / library name'''
+        return self.get_any(self.qualified_name_from_hash(hsh), 'bin')
+
 
 if __name__ == '__main__':
     import argparse
